/*
 *	audiodrv.c
 *
 *  Copyright (C) 1997,1998 Hermann Samso
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "delfin.h"
#include "audiodrv.h"

CHAR* audiodev = NULL;
INT audiobufsize = 128;		//in SAMPLEs
INT freq = 44100;
INT bits = 16;
INT stereo = 1;

CHAR* output = NULL;		//for dumping audio to file
FILE* fp = NULL;


SWORD* audiobuffer = NULL;
INT dspblksize;				//in SAMPLEs
INT audiofd;
BOOL inited = FALSE;
BOOL opened = FALSE;

/*
 * Set sound device parameters for output with 'frequency' in Hz,
 * 'samplesize' in bits, number of 'channels' (0=mono, 1=stereo)
 * and 'buffersize' in SAMPLEs
 *
 */
void AudioInit(CHAR* dev, INT frequency, INT samplesize, INT channels, INT buffersize) {
	if (trace_on)
		fprintf(stderr, "AudioInit()\n");


	audiodev = dev;
	freq = frequency;
	bits = samplesize;
	stereo = channels;
	audiobufsize = buffersize;

	inited = TRUE;
}

/*
 * Open sound device. 'AudioInit' should have been called first
 * If successful return TRUE, else FALSE
 *
 */
BOOL AudioOpen(void) {
	if (trace_on)
		fprintf(stderr, "AudioOpen()\n");


	if (inited == FALSE)
		return(FALSE);
	if (opened)
		return(TRUE);
		
	if ((audiofd = open(audiodev, O_WRONLY, 0)) < 0) {
		fprintf(stderr, "Error: couldn't open sound device: %s\n", audiodev);
		return(FALSE);
	}
	if (ioctl(audiofd, SNDCTL_DSP_SAMPLESIZE, &bits) < 0) {
		fprintf(stderr, "Error: setting data bit size: %d\n", bits);
		close(audiofd);
		return(FALSE);
	}
	if (ioctl(audiofd, SNDCTL_DSP_STEREO, &stereo) < 0) {
		fprintf(stderr, "Error: setting channels: %d\n", stereo);
		close(audiofd);
		return(FALSE);
	}
	if (ioctl(audiofd, SNDCTL_DSP_SPEED, &freq) < 0) {
		fprintf(stderr, "Error: setting sampling rate: %d\n", freq);
		close(audiofd);
		return(FALSE);
	}
	ioctl(audiofd, SNDCTL_DSP_GETBLKSIZE, &dspblksize);
	if (dspblksize < 1024 || dspblksize > (2*65536)) {
		fprintf(stderr, "Error: invalid audio buffer size\n");
		close(audiofd);
		return(FALSE);
	}
	dspblksize = dspblksize/4;	//convert to size in SAMPLEs
	if (AudioSetBufferSize(audiobufsize) == 0)
		return(FALSE);

	opened = TRUE;
	return(TRUE);
}


/*
 * close sound device and free sound buffer memory
 *
 */
void AudioClose(void) {
	if (trace_on)
		fprintf(stderr, "AudioClose()\n");
		
	if (opened == FALSE)
		return;
	if (close(audiofd))
		fprintf(stderr, "Error: closing audio device=%d\n", audiofd);
	free(audiobuffer);
	opened = FALSE;
}


BOOL AudioOpened(void) {
	return(opened);
}


/*
 * check that sound device exists and is writable
 *
 */
BOOL AudioExists(void) {
	if (trace_on)
		fprintf(stderr, "AudioExists()\n");

	if (audiodev == NULL)
		return(FALSE);
	return(access(audiodev, W_OK)==0);
}
	

/*
 * write 'size' SAMPLEs from audiobuffer to audio device
 *
 */
void AudioWrite(ULONG size) {
	ULONG written;

	
	if (trace_on)
		fprintf(stderr, "AudioWrite(%li)\n", size);


//fprintf(stderr,"%d %d %d\n", audiobuffer[0], audiobuffer[8], audiobuffer[16]);

	if (inited == FALSE)
		return;
	if (opened == FALSE)
		return;

	if (size > audiobufsize)	//shouldn't happen, but does it matter?
		size = audiobufsize;
	written = write(audiofd, audiobuffer, size*4);
	if (written != size*4)
		fprintf(stderr, "Error %li: writing samples to audio device\n", written);

//fprintf(stderr, "%d %d %d %d ", audiobuffer[0], audiobuffer[2], audiobuffer[8], audiobuffer[16]);
	// dump audio to file?
	if (output != NULL) {
		if (fp == NULL)
			fp = fopen(output, "a");
		if (fp == NULL)
			return;
		fwrite(audiobuffer, 2, size*2, fp);
		fflush(fp);
	}
}


                                	
BOOL AudioReset(void) {
	if (trace_on)
		fprintf(stderr, "AudioReset()\n");

		
	return(ioctl(audiofd, SNDCTL_DSP_RESET));
}


/*
 * Dump audio to file
 *
 */
void AudioDumpToFile(CHAR* fname) {
	if (trace_on)
		fprintf(stderr, "AudioDumpToFile(%s)\n", fname);

	output = fname;
}


void AudioDumpStop(void) {
	if (trace_on)
		fprintf(stderr, "AudioDumpStop()\n");
	
	if (output == NULL)
		return;
	output = NULL;
}


/*
 * returns size of audiobuffer in SAMPLEs
 *
 */
ULONG AudioGetBufferSize(void) {
	if (trace_on)
		fprintf(stderr, "AudioGetBufferSize()\n");


	return(audiobufsize);
}


SWORD* AudioGetBuffer(void) {
	if (trace_on)
		fprintf(stderr, "AudioGetBuffer()\n");
		
		
	return(audiobuffer);
}



/*
 * Allocate 'audiobuffer' in SWORDs. Size in number of SAMPLES.
 * Returns actually allocated buffer size.
 *
 */
ULONG AudioSetBufferSize(ULONG size) {
	SWORD* buffer = NULL;


	if (trace_on)
		fprintf(stderr, "AudioSetBuffer(%li)", size);
		
	// we need 'dspblksize'... inited?
	if (inited == FALSE)
		return(0);
		
	//each SAMPLE is 2 SWORDs
	buffer = (SWORD*) malloc(size*2*sizeof(SWORD));
	
	if (buffer == NULL) {
		fprintf(stderr, "Error: allocating audio buffer\n");
		return(FALSE);
	}

	// size of buffer can not surpass 'dspblksize'
	if (size > dspblksize)
		audiobufsize = dspblksize;
	else
		audiobufsize = size;
	audiobuffer = buffer;

	return(audiobufsize);
}

